--------------------------Crypto10.e  (i.e.:v1.0)
--This is an encryption library which uses a loose interpretation of the
--"Chain Block Cipher Mode" algorithm, and many twists and turns
--to help obfuscate the resultant cipher and provide a cipher
--that, along with this algorithm, would be very hard to
--cryptoanalize to a smaller set of possibilities and would instead
--require brute forcing (only) as a method of solving.
--(well, dictionary attacks and social engineering are of course
-- potentially viable attacks, as always... that's not really what
-- I am referring to here tho...)

--Now some might say that 32bit isn't very strong, but, I don't
--really think that's true anymore, after studying cryptoanalysis
--for the (yes, very short) time I have, and talking with people
--about it, I feel that what you do to the data as a whole,
--not once you begin xor-ing it (which is what most of the algorithms
--all do anyway, at the end), is what counts. Furthermore, the
--library _may_ only be working on 8bit data :), technically.

--How obfuscated the data already is, before the last step, seems to
--me would matter far more than how many bits at once you are sending
--to be xor'd.

--Now, what, to me, would be *far* stronger than even hundreds
--of bits being processed at a time, like 160bit or whatever, is
--*random* numbers of bits being sent to be xor'd each time. :)
--That would seem... nearly evil :) and since, I've been known,
--on occasion, not to be nice :>, I may incorporate this technique
--into these algorithms at a later time...

--Due to the low "bit strength level" rating, i.e. 32, this library
--should be easily exportable from the U.S. without restriction.

--Also, I, Hawke' <mdeland@geocities.com> freely give this library
--to the public domain, for anyone needing *moderate* strength
--encryption, for what ever reason they may need it. This is provided
--this file is *not* altered in *any* manner, *whatsoever*.

--The single exception to this provision is this include file may
--be compressed for distribution by such compressors as Pkzip, Arj,
--Gzip, Rar, and the like, then distributed/transmitted, and once
--the include file is then de-compressed, it is once again, *not*
--alterable and the provision then becomes effective again.

--If you wish to add functionality to this library, include this
--library into your own library which would contain your enhancements.

--The reason I am enforcing this no modification provision, which
--I feel is a simple provision for a free library, is because I wish
--to ensure that *all* copies of Crypto###.e are *identical*.
--The reason for _that_ is:
--  suppose you implemented a program that depended on the
--  encryption provided by this include/library.  Further, suppose
--  (may the maker forbid) you had a file system error (bad sector,
--  HD failure, lost cluster from power failure... etc) and your copy
--  of this library was corrupted.  Well, you would then need, nearly
--  immediately I imagine, a replacement copy of your version of this
--  library (like, CryptoV12.e or whatever).  If modified versions of
--  this library were propagated hither, dither and yon, you could never
--  be sure that the copy you just downloaded is the right copy until
--  you downloaded it, decompressed if need be, and attempted to run
--  your program again and subsequently found that your passwords still
--  don't work... I can think of many instances where this irritation,
--  and the time that it eats up, are simply not tolerable.  It's far
--  better, in such harrowing times of trying to recover from file
--  system corruption, to *know* that you just got the right version
--  of the library, for sure, and you can go on about finding and
--  retrieving any other files/programs/wotnot that also got thrashed
--  in your file system disaster.
--So, please, do *not*, alter the contents of this file; it's all I ask.


--DISCLAIMER: The library is provided *without* any implied or
--expressed warranty or expression of fitness or merchantability
--towards any application, use, or data transformation, in any
--way, what so ever.  Furthermore, it would be potentially illegal
--to use this library for encryption of certain data, as certain
--types/forms of data *require* specific use of specific algorithms.
--Lastly, this library should **NOT** be used for encryption of
--truly sensitive information, even if it provides a theoretically
--stronger encryption strength than another encryption package that
--is deemed actually useable for the data you are seeking to encrypt.
--If the information is truly sensitive, I would recommend Twofish,
--applied at a bit strength rating deemed appropriate.
--In short: Use at your own risk for stuff that doesn't count. :)

--WARNING:Before you ask, no :), I can not recover keys from ciphered
--text/files/wotnot made by this library.  I'm not sure that any other
--method besides brute-forcing with large computers could.

--Example: A machine that could attempt 1 million potential,
--sequentially generated keys per second, would take (by my calculations)
--something like 3.17e26 *years* to try all keys.  This is arrived by
--the theory that there are at least 1e40 potential keys
--(255^20=1.35e48) after ruling out at least 100 *million* keys...
--(well I know it was not/can not be: "blah" or whatever)
--even if cryptoanalysis and rule-outs brought you down to 1e20
--potential keys remaining to scroll thru, the resultant time, at
--1 million attempts per second is over 3 million years!!!

--For comparison, using this library on my p200MMX, 48Mram, EU2,
--everything cached, I can accomplish 10000 keys tried per second....

--The file encryption algorithms are quite fast, far faster than
--(or, at least *should* be far faster than) algorithms like PGP,
--DES, and the like.  I encrypted/decrypted win32lib.ew V15b,
--which has an approx. size of ~200000 bytes, and ~5500 lines
--on my machine in 0.10-0.15 seconds on average.  This means
--the algorithms can likely maintain speeds of ~1meg of data/sec
--and/or ~25000+ lines/sec... perhaps far more than that...
--(on my machine...) not shabby :)

--It is 'fun' to look at the resultant encrypted file, and you may
--note certain repeated strings within the encrypted text.
--heh... I'm not real sure why this happens, and it's only when
--viewing the encrypted file/data with certain viewers that it does.
--These repeated strings are *highly* likely to *not* be the
--same unencrypted text :) and if you were to attempt something
--like a substitution cryptoanalysis upon them, it
--will likely lead to vast quantities of hair-pulling :)

include machine.e

function makeseed(sequence d)
--given a 'string', generate a unique number based upon that
--string, ranging from 1 to over 1,000,000,000.
integer s s = 0
   for i = 1 to length(d) do
       s = s + ( d[i] * power(2,i+1) )
   end for
   return s 
end function

function reverse_this(sequence d)
--Hi Rob! :)
integer lower, n
sequence t
   n = length(d)
   t = repeat(0, n)
   lower = 1
   for upper = n to floor(n/2)+1 by -1 do
      t[upper] = d[lower]
      t[lower] = d[upper]
      lower = lower + 1
   end for
   return t
end function

function encrline(object line, sequence cipher)
--given a line, and a premade cipher, encrypt the string against
--the cipher, but make sure the cipher isn't 'saved' or 'visible'...
--called from EncryptFile() after making the cipher, to help
--prevent needing to constantly reencrypt the same key, and from
--EncryptSTR, since it's the same code needed after making the cipher
--side note: you gotta love EU... isn't this an elegant function?
integer lenL, lenC
   lenC = length(cipher)
   if lenC = 0 then return {} end if
   if atom(line) then
      return xor_bits(line,cipher[rand(lenC)])
   end if
   lenL = length(line)
   if lenL <= lenC then
        return xor_bits(line,cipher[1..lenL])
   else return encrline(line,cipher & cipher)
   end if
end function

function decrline(object line, sequence cipher)
   return encrline(line,cipher)
end function

global function Encrypt(sequence k)
--given a plaintext key, return a cipher based upon that key, that
--has padding that ensures the cipher won't be likely to reveal the
--original length of the key, might/might not be backwards, is
--heavily randomized, and finally, chain block cipher mode encrypted
--against itself in character succession. 
   if length(k) >20 then puts(1,"Encrypt:key too long") abort(1) end if

   --generate a seed
   set_rand( makeseed(k) )

   --add random padding
   for i = 1 to rand(3) do
       k = k & k[1..rand( length(k) )]
   end for

   --another curve or 2 to help defeat cryptoanalysis and
   --require brute forcing to solve a cipher :)
   if rand(2) = 1 then k = reverse_this(k) end if
   k = rand(k)

   --xor encrypt the string against a seeded, random, version of
   --itself, using a 32bit "Chain Block Cipher Mode" algorithm.
   for i = 1 to length(k)-1 do
       k[i] = xor_bits( k[i], rand(k[i+1]) )
   end for
   return k
end function

global function Decrypt(sequence cipher, sequence key)
--useage: given an encrypted 'cipher' text string and a test case
--  'key' to decrypt by, return t/f if it was/wasn't the right key
   return compare( cipher, Encrypt(key) ) = 0 
end function

global function EncryptSTR(sequence str, sequence key)
--given a string, encrypt it against a key
  return encrline( str, Encrypt(key) )
end function

global function DecryptSTR(sequence str, sequence key)
--given an encrypted string, attempt a decryption based
--upon the attempted key, and return garbage if it was the
--wrong key or the plaintext decryption if it was the right key.
  return encrline( str, Encrypt(key) )
end function

global function EncryptData(sequence data, sequence key)
--given a text file or sequence, and a key, encrypt the data
--based upon that key, with varying cipher methodology applied
--on successive lines within the data.
sequence cipher
   cipher = Encrypt(key)
   for i = 1 to length(cipher) do
      if cipher[i] = 0 then cipher[i] = 255 end if
   end for
   for i = 1 to length(data) do
      data[i] = encrline(data[i],rand(cipher))
   end for
   return data
end function

global function DecryptData(sequence data, sequence key)
--given ciphered data, and an attempted key (right or wrong),
--return a sequence of decrypted text/data based upon applying
--the key to the cipher.  If the wrong key is given, garbage
--will be returned, and if the right key is given, the unencrypted
--file/data will be returned, restored in it's full glory.
   return EncryptData(data,key)
end function

